/*
 * FILE:  fsm73xxs.c
 *
 * WHO:   brian harper / gifford scott
 *
 * DESCR: Define flash chip mapping for the Netgear FSM7328/52S boxes
 *
 * NOTE:  - MTD ==> 'Memory Technology Device'
 *                  (I guess the word 'flash' must be copyrighted ...)
 */

/* assert partition support */
#define HAVE_PARTITIONS

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <asm/io.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/config.h>
#ifdef HAVE_PARTITIONS
#include <linux/mtd/partitions.h>
#endif

__u8 fsm73xxs_read8(struct map_info *map, unsigned long ofs);
__u16 fsm73xxs_read16(struct map_info *map, unsigned long ofs);
__u32 fsm73xxs_read32(struct map_info *map, unsigned long ofs);
void fsm73xxs_copy_from(struct map_info *map, void *to, unsigned long from, ssize_t len);
void fsm73xxs_write8(struct map_info *map, __u8 d, unsigned long adr);
void fsm73xxs_write16(struct map_info *map, __u16 d, unsigned long adr);
void fsm73xxs_write32(struct map_info *map, __u32 d, unsigned long adr);
void fsm73xxs_copy_to(struct map_info *map, unsigned long to, const void *from, ssize_t len);


/* define flash physical address origin */
#define FLASH_PHY_ORG           0x1c000000L


#ifdef HAVE_PARTITIONS
static struct mtd_partition fsm73xxs_partitions[] =
{
  /* partitionName              size        offset      maskFlags */
	{"U-Boot",                    0x00020000, 0x00000000, 0},
	{"U-Boot Environment",        0x00020000, 0x00020000, 0},
	{"Application Filesystem",    0x00DC0000, 0x00040000, 0},
	{"System (Kernel + Ramdisk)", 0x00200000, 0x00E00000, 0}
};

#define NUM_PARTITIONS  (sizeof(fsm73xxs_partitions)/sizeof(struct mtd_partition))

/* MTD device map - see description in map.h */
struct map_info fsm73xxs_map =
{
  /* mapName                    size        busWidth (in bytes) */
  "Netgear FSM73xxS flash",     0x01000000, 2,
  fsm73xxs_read8,
  fsm73xxs_read16,
  fsm73xxs_read32,
  NULL, NULL,           /* read64, write64 */
  fsm73xxs_copy_from,
  fsm73xxs_write8,
  fsm73xxs_write16,
  fsm73xxs_write32,
  fsm73xxs_copy_to
};
#endif

static struct mtd_info *mymtd;


__u8 fsm73xxs_read8(struct map_info *map, unsigned long offset)
{
	return __raw_readb(map->map_priv_1 + offset);
}

__u16 fsm73xxs_read16(struct map_info *map, unsigned long offset)
{
	return __raw_readw(map->map_priv_1 + offset);
}

__u32 fsm73xxs_read32(struct map_info *map, unsigned long offset)
{
	return __raw_readl(map->map_priv_1 + offset);
}

void fsm73xxs_copy_from(struct map_info *map, void *to, unsigned long from, ssize_t len)
{
  /* for copying from flash, just do memcpy() */
	memcpy_fromio(to, map->map_priv_1 + from, len);
}

void fsm73xxs_write8(struct map_info *map, __u8 d, unsigned long offset)
{
	__raw_writeb(d, map->map_priv_1 + offset);
	mb();
}

void fsm73xxs_write16(struct map_info *map, __u16 d, unsigned long offset)
{
	__raw_writew(d, map->map_priv_1 + offset);
	mb();
}

void fsm73xxs_write32(struct map_info *map, __u32 d, unsigned long offset)
{
	__raw_writel(d, map->map_priv_1 + offset);
	mb();
}

void fsm73xxs_copy_to(struct map_info *map, unsigned long to, const void *from, ssize_t len)
{
	memcpy_toio(map->map_priv_1 + to, from, len);
}


int __init fsm73xxsMapInit(void)
{
  printk( KERN_NOTICE "%s device: 0x%lX bytes @ phy addr 0x%lX\n\r",
          fsm73xxs_map.name, fsm73xxs_map.size, FLASH_PHY_ORG );

	fsm73xxs_map.map_priv_1 = (unsigned long) ioremap( FLASH_PHY_ORG,
                                                     fsm73xxs_map.size );
	if (fsm73xxs_map.map_priv_1 == (unsigned long) NULL)
  {
		printk("Failed to ioremap\n");
		return -EIO;
	}
  else
  {
		printk( "Mapped to virt addr 0x%lX\n\r", fsm73xxs_map.map_priv_1 );
  }

	mymtd = do_map_probe("cfi_probe", &fsm73xxs_map);
	if (mymtd == NULL)
  {
    iounmap((void *)fsm73xxs_map.map_priv_1);
    return -ENXIO;
	}

  mymtd->module = THIS_MODULE;

#if 0
  /* TEST */
  {
    typedef unsigned int uint32_t;
    #define EVER                  ;;
    #define STRINGIZE(X)          #X
    #define BYTE_COUNT            16
    #define OFFSET                0x30100
    #define TEST_INIT(OFFSET, BYTE_COUNT) "fsm73xxsMapInit: write " STRINGIZE(BYTE_COUNT) \
                                  " chars to offset " STRINGIZE(OFFSET) "\n\r"
    #define TEST_COMPLETE(VAR)    "fsm73xxsMapInit: write complete, " STRINGIZE(VAR) \
                                  " = 0x%X\n\r", VAR
    
    char txtDeadbeef[BYTE_COUNT];
    char* pBfr = txtDeadbeef;
    char* const pBfrOrg = txtDeadbeef;
    char* const pBfrTerm = txtDeadbeef + BYTE_COUNT;
    uint32_t retByteCount;

    /* init deadbeef */
    while (pBfr < pBfrTerm)
    {
      *pBfr = (char) (uint32_t) (void*) (pBfr - pBfrOrg);

      /* adjust loop var */
      pBfr += 1;
    }

    printk( TEST_INIT(OFFSET, BYTE_COUNT) );
    mymtd->write(mymtd, OFFSET, BYTE_COUNT, &retByteCount, txtDeadbeef);
    printk( TEST_COMPLETE(retByteCount) );

    printk( TEST_INIT(OFFSET + 0x40, 8) );
    mymtd->write(mymtd, OFFSET + 0x40, 8, &retByteCount, txtDeadbeef);
    printk( TEST_COMPLETE(retByteCount) );

    printk( TEST_INIT(OFFSET + 0x80, 4) );
    mymtd->write(mymtd, OFFSET + 0x80, 4, &retByteCount, txtDeadbeef);
    printk( TEST_COMPLETE(retByteCount) );

    printk( TEST_INIT(OFFSET + 0xC0, 2) );
    mymtd->write(mymtd, OFFSET + 0xC0, 2, &retByteCount, txtDeadbeef);
    printk( TEST_COMPLETE(retByteCount) );

    for (EVER);
  }
#endif

#if defined(HAVE_PARTITIONS)
  add_mtd_partitions(mymtd, fsm73xxs_partitions, NUM_PARTITIONS);
#else
  add_mtd_device(mymtd);
#endif

  return 0;
}


static void __exit fsm73xxsMapExit(void)
{
  /* clean-up partitions */
	if (mymtd != NULL)
  {
#if defined(HAVE_PARTITIONS)
		del_mtd_partitions(mymtd);
#else
		del_mtd_device(mymtd);
#endif
		map_destroy(mymtd);
	}

  /* unmap */
	if (fsm73xxs_map.map_priv_1)
  {
		iounmap((void *) fsm73xxs_map.map_priv_1);
		fsm73xxs_map.map_priv_1 = 0;
	}
}

module_init(fsm73xxsMapInit);
module_exit(fsm73xxsMapExit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("brian W. harper");
MODULE_DESCRIPTION("MTD map driver for Netgear FSM73xxS");
